# -*- coding: utf-8 -*-
#
# satextrato/__init__.py
#
# Copyright 2015 Base4 Sistemas Ltda ME
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import
from __future__ import unicode_literals

import logging


__version__ = '0.3'

logging.getLogger('satextrato').addHandler(logging.NullHandler())


from .venda import ExtratoCFeVenda  # noqa: E402, F401
from .cancelamento import ExtratoCFeCancelamento  # noqa: E402, F401



# -*- coding: utf-8 -*-
#
# tests/conftest.py
#
# Copyright 2015 Base4 Sistemas Ltda ME
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import importlib
import os
import shutil

from builtins import str as text

import pytest

from escpos import constants as escpos_const
from escpos.conn import CONNECTION_TYPES


def pytest_addoption(parser):

    parser.addoption(
            '--escpos-impl',
            action='store',
            metavar='CLASSNAME',
            default='escpos.impl.epson.GenericESCPOS',
            help='Implementacao ESC/POS a ser instanciada'
        )

    parser.addoption(
            '--escpos-if',
            action='store',
            default='dummy',
            choices=[alias for alias, type_info in CONNECTION_TYPES],
            help='Interface com o dispositivo'
        )

    parser.addoption(
            '--escpos-if-settings',
            action='store',
            metavar='SETTINGS',
            default='',
            help='String de conexão (conforme a interface com o dispositivo)'
        )

    parser.addoption(
            '--escpos-encoding',
            action='store',
            metavar='ENCODING',
            default=escpos_const.DEFAULT_ENCODING,
            help='Encoding antes de enviar dados ao dispositivo'
        )

    parser.addoption(
            '--escpos-encoding-errors',
            action='store',
            metavar='ERRORS',
            default=escpos_const.DEFAULT_ENCODING_ERRORS,
            help='Como lidar com erros de codificação de caracteres'
        )


class InterfaceFactory(object):

    def __init__(self, request):
        self._request = request

    def get_connection_interface(self):
        interface = self._request.config.getoption('--escpos-if')
        settings = self._request.config.getoption('--escpos-if-settings')
        method = 'create_{}_connection'.format(interface)
        return getattr(self, method)(settings)

    def create_bluetooth_connection(self, settings):
        from escpos import BluetoothConnection
        conn = BluetoothConnection.create(settings)
        return conn

    def create_dummy_connection(self, settings):
        from escpos import DummyConnection
        conn = DummyConnection.create(settings)
        return conn

    def create_file_connection(self, settings):
        from escpos import FileConnection
        conn = FileConnection.create(settings)
        return conn

    def create_network_connection(self, settings):
        from escpos import NetworkConnection
        conn = NetworkConnection.create(settings)
        return conn

    def create_serial_connection(self, settings):
        from escpos import SerialConnection
        conn = SerialConnection.create(settings)
        return conn

    def create_usb_connection(self, settings):
        from escpos import USBConnection
        conn = USBConnection.create(settings)
        return conn


@pytest.fixture(scope='function')
def datadir(tmpdir, request):
    path, _ = os.path.split(request.module.__file__)
    dirname = os.path.join(path, 'data')
    shutil.copytree(dirname, text(tmpdir.join('data')))
    return tmpdir


@pytest.fixture(scope='session')
def configuracao(request):
    from satextrato import config
    conf = config.padrao()
    return conf


@pytest.fixture(scope='session')
def escpos_impl(request):
    names = request.config.getoption('--escpos-impl').split('.')
    _module = importlib.import_module('.'.join(names[:-1]))
    printer = getattr(_module, names[-1])

    encoding = request.config.getoption('--escpos-encoding')
    encoding_errors = request.config.getoption('--escpos-encoding-errors')

    factory = InterfaceFactory(request)
    device = factory.get_connection_interface()

    impl = printer(
            device,
            encoding=encoding,
            encoding_errors=encoding_errors
        )
    return impl



# -*- coding: utf-8 -*-
#
# satextrato/base.py
#
# Copyright 2015 Base4 Sistemas Ltda ME
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import textwrap
import xml.etree.ElementTree as ET

from six.moves import range

from unidecode import unidecode

import escpos.barcode
import escpos.feature

from satcomum import br
from satcomum import constantes

from .config import padrao as config_padrao


class ExtratoCFe(object):
    """Classe base para os extratos do CF-e de venda e cancelamento, fornecendo
    uma implementação comum para o cabeçalho dos CF-e de venda e cancelamento,
    além da infraestrutura que simplifica a interface para impressoras ESC/POS.

    As implementações que realmente emitem os extratos estão nas classes
    :class:`~satextrato.venda.ExtratoCFeVenda` e
    :class:`~satextrato.cancelamento.ExtratoCFeCancelamento`.
    """

    def __init__(self, fp, impressora, config=None):
        """Inicia uma instância de :class:`ExtratoCFe`.

        :param fp: Um *file object* para o documento XML que contém o CF-e.

        :param impressora: Um instância (ou subclasse, especialização) de
            :class:`escpos.impl.epson.GenericESCPOS` usado para efetivamente
            imprimir o extrato.

        :param config: Opcional.
            Uma instância de :class:`satextrato.config.Configuracoes`.
            Se não informado, serão usados valores padrão.

        """
        super(ExtratoCFe, self).__init__()
        self._config = config or config_padrao()
        self._tree = ET.parse(fp)
        self.root = self._tree.getroot()
        self.impressora = impressora

        self._flag_negrito = False
        self._flag_italico = False
        self._flag_expandido = False
        self._flag_condensado = False

    @property
    def _colunas(self):
        if self._flag_condensado:
            num_colunas = self.impressora.feature.columns.condensed
        elif self._flag_expandido:
            num_colunas = self.impressora.feature.columns.expanded
        else:
            num_colunas = self.impressora.feature.columns.normal
        return num_colunas

    @property
    def is_ambiente_testes(self):
        """Indica se o CF-e-SAT foi emitido em "ambiente de testes".
        Considera como emitido em ambiente de testes quando:

        * Elemento B10 ``tpAmb`` for ``2`` (ambiente de testes) ou
        * Elemento B12 ``signAC`` possuir a assinatura de teste, indicada pela
          constante :attr:`satcomum.constantes.ASSINATURA_AC_TESTE`.

        .. note::

            O CF-e de cancelamento não possui o elemento ``tpAmb``, conforme
            descrito na ER SAT, item 4.2.3 **Layout do CF-e de cancelamento**.

        :raises ValueError: Se o documento XML não for identificado como um
            CF-e-SAT de venda ou cancelamento.

        """
        signAC = self.root.findtext('./infCFe/ide/signAC')

        if self.root.tag == constantes.ROOT_TAG_VENDA:
            tpAmb = self.root.findtext('./infCFe/ide/tpAmb')
            return (tpAmb == constantes.B10_TESTES
                    or signAC == constantes.ASSINATURA_AC_TESTE)

        elif self.root.tag == constantes.ROOT_TAG_CANCELAMENTO:
            # CF-e-SAT de cancelamento não possui `tpAmb`
            return signAC == constantes.ASSINATURA_AC_TESTE

        raise ValueError(
                (
                    'Documento nao parece ser um CF-e-SAT; root tag: {!r}'
                ).format(self.root.tag)
            )

    def imprimir(self):
        self.cabecalho()
        self.corpo()
        self.rodape()
        self.fim_documento()

    def centro(self):
        self.impressora.justify_center()
        return self

    def esquerda(self):
        self.impressora.justify_left()
        return self

    def normal(self):
        if self._flag_negrito:
            self.negrito()
        if self._flag_italico:
            self.italico()
        if self._flag_expandido:
            self.expandido()
        if self._flag_condensado:
            self.condensado()

    def negrito(self):
        self._flag_negrito = not self._flag_negrito
        self.impressora.set_emphasized(self._flag_negrito)
        return self

    def italico(self):
        self._flag_italico = not self._flag_italico
        return self

    def expandido(self):
        self._flag_expandido = not self._flag_expandido
        self.impressora.set_expanded(self._flag_expandido)
        return self

    def condensado(self):
        self._flag_condensado = not self._flag_condensado
        self.impressora.set_condensed(self._flag_condensado)
        return self

    def bordas(
            self,
            texto_esquerda,
            texto_direita,
            colunas=None,
            espacamento_minimo=4
            ):
        largura = colunas or self._colunas
        texto = _bordas(
                texto_esquerda,
                texto_direita,
                largura=largura,
                espacamento_minimo=espacamento_minimo
            )
        self.texto(texto)
        return self

    def quebrar(self, texto, colunas=None):
        largura = colunas or self._colunas
        # considera hard-breaks
        linhas_fixas = texto.replace('\r', '').split('\n')
        for linha_fixa in linhas_fixas:
            linhas = textwrap.wrap(linha_fixa, largura)
            for linha in linhas:
                self.texto(linha)
        return self

    def avanco(self, linhas=1):
        self.impressora.lf(lines=linhas)
        return self

    def texto(self, texto):
        self.impressora.text(unidecode(texto))
        return self

    def separador(self, caracter='-', colunas=None):
        largura = colunas or self._colunas
        self.texto('-' * largura)
        return self

    def indicacao_de_teste(self):
        """Imprime indicação de teste se o CF-e estiver em "condição de
        teste". Caso contrário, não faz nada. A indicação de teste,
        conforme o Manual de Orientação deverá ser impressa em itálico
        (note a linha em branco acima e abaixo da inscrição "TESTE"):

        .. sourcecode:: text

                    10        20        30        40      48 :
            ....:....|....:....|....:....|....:....|....:..! :
                                                             :
                             = T E S T E =                   :
                                                             :
            <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< :
            <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< :
            <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< :

        .. note::

            O correto são linhas com sinais de ``>``, mas isso faz com que o
            framework de teste interprete como um bloco de teste (doctest). Os
            sinais de ``<`` dão uma boa ideia do resultado final.

        """
        if self.is_ambiente_testes:
            self.italico()
            self.avanco()
            self.texto('= T E S T E =')
            self.avanco()
            for i in range(3):
                self.texto('>' * self._colunas)
            self.italico()

    def numero_extrato(self):
        """
        Obtém o número do extrato, elemento ``nCFe`` (B06). Se o CF-e estiver
        em condição de teste retornará ``000000``.

        Veja :meth:`is_ambiente_testes` para outros detalhes.
        """
        if self.is_ambiente_testes:
            return '0' * 6
        return self.root.findtext('./infCFe/ide/nCFe')

    def chave_cfe_code128(self, chave):
        """Imprime o código de barras padrão Code128 para a chave do CF-e.

        :param chave: Instância de :class:`satcomum.ersat.ChaveCFeSAT`.
        """
        if self._config.code128.ignorar:
            return

        code128_params = dict(
                barcode_height=self._config.code128.altura,
                barcode_width=escpos.barcode.BARCODE_NORMAL_WIDTH,
                barcode_hri=escpos.barcode.BARCODE_HRI_NONE
            )

        if self._config.code128.truncar:
            tamanho = self._config.code128.truncar_tamanho
            digitos = ''.join(chave.partes())[:tamanho]
            self.impressora.code128(digitos, **code128_params)

        elif self._config.code128.quebrar:
            partes = _quebrar_chave(chave, self._config.code128.quebrar_partes)
            for n_parte, parte in enumerate(partes, 1):
                self.impressora.code128(parte, **code128_params)
                if n_parte < len(partes):
                    self.avanco()
        else:
            partes = chave.partes(1)
            self.impressora.code128(partes[0], **code128_params)

    def qrcode_mensagem(self):
        mensagem = self._config.qrcode.mensagem.strip()
        if not mensagem:
            return

        if self._config.qrcode.mensagem_modo_condensado:
            self.condensado()
        self.centro()
        for linha in textwrap.wrap(mensagem, self._colunas):
            self.texto(linha)
        self.esquerda()
        if self._config.qrcode.mensagem_modo_condensado:
            self.condensado()

    def fim_documento(self):
        """Encerra o documento, imprimindo o rodapé (se houver) e avançando ou
        guilhotinando o documento, conforme as configurações.
        """
        self.normal()
        self.avanco()
        self.separador()

        conf_cupom = self._config.cupom
        conf_rodape = self._config.rodape

        if conf_rodape.esquerda or conf_rodape.direita:
            self.condensado()
            self.bordas(conf_rodape.esquerda, conf_rodape.direita)
            self.condensado()

        if self.impressora.feature.cutter and conf_cupom.cortar_documento:
            if conf_cupom.avancar_linhas > 0:
                self.avanco(conf_cupom.avancar_linhas)
            self.impressora.cut(
                    partial=conf_cupom.cortar_parcialmente,
                    feed=conf_cupom.cortar_avanco
                )
        else:
            # impressora não possui guilhotina ou não é para cortar o documento
            if conf_cupom.avancar_linhas > 0:
                self.avanco(conf_cupom.avancar_linhas)

    def cabecalho(self):

        self.normal()

        emit = self.root.find('./infCFe/emit')
        enderEmit = emit.find('enderEmit')

        nome_fantasia = emit.findtext('xFant')
        razao_social = emit.findtext('xNome')

        logradouro = enderEmit.findtext('xLgr')
        numero = enderEmit.findtext('nro')
        complemento = enderEmit.findtext('xCpl')
        bairro = enderEmit.findtext('xBairro')

        if numero:  # número não é obrigatório
            logradouro = u'{}, {}'.format(logradouro, numero)

        if complemento:  # complemento não é obrigatório
            # faz um esforço para manter o complemento na mesma linha que o
            # logradouro/número se couber, senão o complemento irá usar uma
            # linha exclusiva...
            if len(logradouro) + len(complemento) < self._colunas:
                # ok, mantém o complemento na mesma linha que o logradouro
                logradouro = u'{}, {}'.format(logradouro, complemento)
                complemento = ''  # ignora a linha que deveria conter o xCpl

        cidade = u'{}/{} CEP: {}'.format(
                enderEmit.findtext('xMun'),
                br.uf_pelo_codigo(int(self.root.findtext('./infCFe/ide/cUF'))),
                br.as_cep(enderEmit.findtext('CEP')))

        partes_endereco = [logradouro, complemento, bairro, cidade]
        endereco = u'\r\n'.join([e for e in partes_endereco if e])

        cnpj = 'CNPJ: {}'.format(br.as_cnpj(emit.findtext('CNPJ')))
        im = 'IM: {}'.format(emit.findtext('IM') or '')
        ie = 'IE: {}'.format(emit.findtext('IE'))

        self.centro()
        self.negrito()

        if nome_fantasia:
            self.quebrar(nome_fantasia)

        self.quebrar(razao_social)
        self.negrito()

        self.quebrar(endereco)

        self.avanco()
        self.esquerda()
        self.texto(cnpj)
        self.texto(ie)

        if im:
            self.texto(im)

        self.separador()

    def rodape(self):
        raise NotImplementedError()

    def corpo(self):
        raise NotImplementedError()


def _bordas(
        esquerda,
        direita,
        largura=48,
        espacamento_minimo=4,
        favorecer_direita=True
        ):
    """Prepara duas strings para serem impressas alinhadas às bordas opostas da
    mídia, os textos da esquerda (borda esquerda) e da direita (borda direita),
    respeitando uma largura e espaçamento mínimo determinados.

    .. sourcecode:: python

        >>> _bordas('a', 'b')
        'a                                              b'

        >>> esquerda = 'a' * 30
        >>> direita = 'b' * 30
        >>> _bordas(esquerda, direita)
        'aaaaaaaaaaaaaaaaaaaaaa    bbbbbbbbbbbbbbbbbbbbbb'

        >>> esquerda = 'Gazeta publica hoje breve nota de faxina na quermesse'
        >>> direita = 'Um pequeno jabuti xereta viu dez cegonhas felizes'
        >>> _bordas(esquerda, direita, espacamento_minimo=1)
        'Gazeta publica hoje bre viu dez cegonhas felizes'


    :param str esquerda: O texto a ser exibido à esquerda. Se o texto não
        couber (em relação à largura e ao espaçamento mínimo) será exibida
        apenas a porção mais à esquerda desse texto, sendo cortados (não
        impressos) os caracteres do final do texto.

    :param str direita: O texto à ser exibido à direita. Se o texto da direita
        não couber (em relação à largura e ao espaçamento mínimo) será exibida
        apenas porção mais à direita desse texto, sendo cortados (não
        impressos) os caracteres do início do texto.

    :param int largura: Largura em caracteres a considerar ao calcular o vão
        entre os textos da esquerda e direita. O padrão é 48, já que é a
        largura mais comum entre as impressoras térmicas de bobina quando
        imprimindo com a fonte normal.

    :param int espacamento_minimo: Opcional. Determina o número de espaços
        mínimo a ser deixado entre os textos da esquerda e direita. O padrão
        são quatro espaços.

    :param bool favorecer_direita: Opcional. Determina se o texto da direita
        deverá ser favorecido com um espaço maior quando houver diferença
        (sobra) entre os textos da esquerda e direita em relação ao espaçamento
        mínimo determinado. O padrão é favorecer o texto da direita, já que é
        normalmente o dado relevante, como um valor ou percentual.

    :returns: Uma string contendo os textos da esquerda e direita encaixados na
        largura determinada, respeitando um espaçamento mínimo entre eles. Se
        necessário os textos serão truncados para respeitar a largura (o texto
        da esquerda será truncado no final e o texto da direita será truncado
        no início).

    :rtype: str

    """
    espacamento = largura - (len(esquerda) + len(direita))
    if espacamento < espacamento_minimo:
        espacamento = espacamento_minimo
        cpmax = int((largura - espacamento) // 2)
        cpmax_esq, cpmax_dir = cpmax, cpmax
        diferenca = largura - (espacamento + cpmax * 2)
        if diferenca > 0:
            if favorecer_direita:
                cpmax_dir += diferenca
            else:
                cpmax_esq += diferenca
        esquerda = esquerda[:cpmax_esq]
        direita = direita[-cpmax_dir:]
    return '%s%s%s' % (esquerda, ' ' * espacamento, direita)


def _quebrar_chave(chave, quebrar_partes):
    # chave: satcomum.ersat.ChaveCFeSAT
    # quebrar_partes: list[int]
    # Lista <quebrar_partes> deve possuir apenas números pares
    digitos = ''.join(chave.partes())
    partes = []
    a = 0
    for n in quebrar_partes:
        partes.append(digitos[a:a+n])
        a += n
    return partes



# -*- coding: utf-8 -*-
#
# satextrato/cancelamento.py
#
# Copyright 2015 Base4 Sistemas Ltda ME
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import
from __future__ import unicode_literals

import xml.etree.ElementTree as ET

from datetime import datetime
from decimal import Decimal

import six

from satcomum import br
from satcomum import ersat

from .base import ExtratoCFe


class ExtratoCFeCancelamento(ExtratoCFe):
    """Implementa impressão do extrato do CF-e de cancelamento."""

    def __init__(self, fp_venda, fp_canc, impressora, config=None):
        """Inicia uma instância de :class:`ExtratoCFeCancelamento`.

        :param fp_venda: Um *file object* para o XML do CF-e de venda.
            Neste caso, o documento XML da venda que foi cancelada.

        :param fp_canc: Um *file object* para o XML do CF-e de cancelamento.

        :param impressora: Um instância (ou subclasse, especialização) de
            :class:`escpos.impl.epson.GenericESCPOS` usado para efetivamente
            imprimir o extrato.

        :param config: Opcional.
            Uma instância de :class:`satextrato.config.Configuracoes`.
            Se não informado, serão usados valores padrão.

        """
        super(ExtratoCFeCancelamento, self).__init__(
                fp_canc,
                impressora,
                config=config
            )
        # (!) `self.root` mantém uma referência para o elemento `CFeCanc`
        self._tree_venda = ET.parse(fp_venda)
        self.root_venda = self._tree_venda.getroot()  # ref. elemento `CFe`

    def corpo(self):
        self.corpo_titulo()
        self.corpo_dados_consumidor()
        self.corpo_total_cupom()
        self.corpo_dados_cfe_cancelado()

    def corpo_titulo(self):
        self.normal()
        self.centro()
        self.negrito()
        self.texto('Extrato No. {}'.format(self.numero_extrato()))
        self.texto(u'CUPOM FISCAL ELETRÔNICO - SAT')
        self.texto(u'CANCELAMENTO')
        self.indicacao_de_teste()
        self.negrito()
        self.esquerda()
        self.separador()
        self.negrito()
        self.condensado()
        self.texto(u'DADOS DO CUPOM FISCAL ELETRÔNICO CANCELADO')
        self.condensado()
        self.negrito()

    def corpo_dados_consumidor(self):
        documento = (
                self.root.findtext('./infCFe/dest/CNPJ')
                or self.root.findtext('./infCFe/dest/CPF')
                or ''
            )

        if br.is_cnpjcpf(documento):
            self.normal()
            self.esquerda()
            self.avanco()
            self.texto('CPF/CNPJ do Consumidor: {}'.format(
                    br.as_cnpjcpf(documento)))
        else:
            # não há um consumidor identificado; apenas normaliza e avança
            self.normal()
            self.avanco()

    def corpo_total_cupom(self):
        self.normal()
        self.esquerda()
        self.negrito()
        self.texto('TOTAL R$ {:n}'.format(
                Decimal(self.root.findtext('./infCFe/total/vCFe'))))
        self.negrito()

    def corpo_dados_cfe_cancelado(self):

        infCFe = self.root_venda.find('./infCFe')  # (!) infCFe da venda
        sat_numero_serie = 'SAT no. {}'.format(
                infCFe.findtext('ide/nserieSAT')
            )

        datahora = datetime.strptime('{}{}'.format(
                infCFe.findtext('ide/dEmi'),
                infCFe.findtext('ide/hEmi')), '%Y%m%d%H%M%S')

        datahora_emissao = datahora.strftime('%d/%m/%Y - %H:%M:%S')
        if six.PY2:
            # Python 2: strftime() resulta em str, que precisará ser passado
            # como um objeto unicode para unidecode (em self.texto())
            datahora_emissao = datahora_emissao.decode('utf-8')

        chave = ersat.ChaveCFeSAT(infCFe.attrib['Id'])

        self.normal()
        self.centro()
        self.avanco()
        self.negrito()
        self.texto(sat_numero_serie)
        self.negrito()
        self.texto(datahora_emissao)

        self.avanco()
        self.esquerda()
        self.condensado()
        self.texto(' '.join(chave.partes()))
        self.condensado()

        self.avanco()
        self.centro()

        self.chave_cfe_code128(chave)

        self.avanco(2)
        self.impressora.qrcode(
                ersat.dados_qrcode(self._tree_venda),
                qrcode_module_size=self._config.qrcode.tamanho_modulo,
                qrcode_ecc_level=self._config.qrcode.nivel_correcao
            )

    def rodape(self):

        infCFe = self.root.find('./infCFe')  # (!) infCFe do cancelamento
        sat_numero_serie = 'SAT no. {}'.format(
                infCFe.findtext('ide/nserieSAT')
            )

        datahora = datetime.strptime('{}{}'.format(
                infCFe.findtext('ide/dEmi'),
                infCFe.findtext('ide/hEmi')), '%Y%m%d%H%M%S')

        datahora_emissao = datahora.strftime('%d/%m/%Y - %H:%M:%S')
        if six.PY2:
            # Python 2: strftime() resulta em str, que precisará ser passado
            # como um objeto unicode para unidecode (em self.texto())
            datahora_emissao = datahora_emissao.decode('utf-8')

        chave = ersat.ChaveCFeSAT(infCFe.attrib['Id'])

        self.normal()
        self.avanco(2)
        self.separador()
        self.centro()
        self.negrito()
        self.quebrar(u'DADOS DO CUPOM FISCAL ELETRÔNICO DE CANCELAMENTO')
        self.avanco()
        self.texto(sat_numero_serie)
        self.negrito()
        self.texto(datahora_emissao)

        self.avanco()
        self.esquerda()
        self.condensado()
        self.texto(' '.join(chave.partes()))
        self.condensado()

        self.avanco()
        self.centro()

        self.chave_cfe_code128(chave)

        self.avanco(2)
        self.impressora.qrcode(
                ersat.dados_qrcode(self._tree),
                qrcode_module_size=self._config.qrcode.tamanho_modulo,
                qrcode_ecc_level=self._config.qrcode.nivel_correcao
            )

        self.avanco()
        self.qrcode_mensagem()



# -*- coding: utf-8 -*-
#
# tests/test_satextrato.py
#
# Copyright 2015 Base4 Sistemas Ltda ME
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

import io

from builtins import str as text

from satextrato.venda import ExtratoCFeVenda
from satextrato.cancelamento import ExtratoCFeCancelamento


def test_extrato_venda(
        configuracao,
        datadir,
        escpos_impl):
    cfe_venda = text(datadir.join('data', 'cfe_venda.xml'))
    with io.open(cfe_venda, 'r') as stream:
        extrato = ExtratoCFeVenda(stream, escpos_impl, config=configuracao)
        extrato.imprimir()


def test_extrato_venda_resumido(
        configuracao,
        datadir,
        escpos_impl):
    cfe_venda = text(datadir.join('data', 'cfe_venda.xml'))
    with io.open(cfe_venda, 'r') as stream:
        extrato = ExtratoCFeVenda(
                stream,
                escpos_impl,
                resumido=True,
                config=configuracao
            )
        extrato.imprimir()


def test_extrato_cancelamento(
        configuracao,
        datadir,
        escpos_impl):
    cfe_venda = text(datadir.join('data', 'cfe_venda.xml'))
    cfe_canc = text(datadir.join('data', 'cfe_cancelamento.xml'))
    with io.open(cfe_venda, 'r') as stream_venda, \
            io.open(cfe_canc, 'r') as stream_canc:
        extrato = ExtratoCFeCancelamento(
                stream_venda,
                stream_canc,
                escpos_impl,
                config=configuracao
            )
        extrato.imprimir()



# -*- coding: utf-8 -*-
#
# satextrato/venda.py
#
# Copyright 2015 Base4 Sistemas Ltda ME
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import
from __future__ import unicode_literals

import textwrap

from datetime import datetime
from decimal import Decimal

import six

from satcomum import br
from satcomum import ersat
from satcomum import util

from .base import _bordas
from .base import ExtratoCFe


ZERO = Decimal()


class ExtratoCFeVenda(ExtratoCFe):
    """Implementa impressão do extrato do CF-e de venda, normal e resumido."""

    def __init__(self, fp, impressora, resumido=False, config=None):
        """Inicia uma instância de :class:`ExtratoCFeVenda`.

        :param fp: Um *file object* para o XML do CF-e de venda.

        :param impressora: Um instância (ou subclasse, especialização) de
            :class:`escpos.impl.epson.GenericESCPOS` usado para efetivamente
            imprimir o extrato.

        :param bool resumido: Opcional. Indica se deverá ser impressa a
            versão resumida do extrato. Se não informado, assume ``False``.

        :param conf: Opcional.
            Uma instância de :class:`satextrato.config.Configuracoes`.
            Se não informado, serão usados valores padrão.

        """
        super(ExtratoCFeVenda, self).__init__(fp, impressora, config=config)
        self._resumido = resumido
        self.anotacoes_antes_obs_contribuinte = []
        self.anotacoes_corpo = []

    def apresentar_item(self, det):
        """Apresenta um item (produto/serviço) do CF-e.

        Nesta abordagem, os detalhes do item utilizam uma linha exclusiva,
        abaixo das linhas que mostram o número do item, código e descrição.
        Por exemplo (considerando 48 caracteres por linha):

        .. sourcecode:: text

                    10        20        30        40      48 :
            ....:....|....:....|....:....|....:....|....:..! :
            001 7891234567891 PELLENTESQUE HABITANT MORBI    :
            TRISTIQUE SENECTUS ET NETUS ET MALESUADA FAMES   :
            AC TURPIS EGESTAS VESTIBULUM TORTOR QUAM AMET    :
                                   1 UN x 12,75 (0,30) 12,75 :

        Mesmo no caso em que o nome do produto é muito pequeno, serão
        utilizadas duas linhas para apresentar o item.

        .. sourcecode:: text

                    10        20        30        40      48 :
            ....:....|....:....|....:....|....:....|....:..! :
            001 7891234567891 AB                             :
                                   1 UN x 12,75 (0,30) 12,75 :

        Mas, como é muito rara a situação em que o produto poderia ser
        apresentado em uma única linha, o ganho na facilidade de leitura
        compensa (pelo menos para o consumidor).

        Este método ainda faz um esforço de apresentar o item em apenas uma
        linha tentando "encaixar" o detalhamento do item na mesma linha de
        descrição, desde que caiba. Isso pode ser mais comum, se estiver
        configurado para apresentar os itens em modo condensado (padrão).

        :param det: O n-ésimo elemento ``/CFe/infCFe/det`` (produto ou
            serviço) existente no corpo do CF-e de venda.

        """
        nItem = int(det.attrib['nItem'])
        prod = det.find('prod')

        cProd = prod.findtext('cProd')
        xProd = prod.findtext('xProd')
        uCom = prod.findtext('uCom')
        qCom = Decimal(prod.findtext('qCom'))
        vUnCom = Decimal(prod.findtext('vUnCom'))
        vProd = Decimal(prod.findtext('vProd'))
        vItem12741 = Decimal(det.findtext('imposto/vItem12741') or 0)

        if vItem12741.is_zero():
            detalhe = u'{:s} {:s} x {:n} {:n}'.format(
                    util.texto_decimal(qCom), uCom, vUnCom, vProd)
        else:
            detalhe = u'{:s} {:s} x {:n} ({:n}) {:n}'.format(
                    util.texto_decimal(qCom), uCom, vUnCom, vItem12741, vProd)

        texto_item = u'{0:03d} {1:s} {2:s}'.format(nItem, cProd, xProd)

        self.normal()
        self.esquerda()

        if self._config.cupom.itens_modo_condensado:
            self.condensado()

        largura = self._colunas
        linhas = textwrap.wrap(texto_item, largura)

        ultima_linha = linhas[-1]

        if (len(ultima_linha) + len(detalhe) + 1) > largura:
            # não cabe, o detalhe será apresentado em uma linha exclusiva
            linhas.append(_bordas(
                    '',
                    detalhe,
                    largura=largura,
                    espacamento_minimo=1
                ))
        else:
            # cabe; o detalhe será apresentado junto com a última linha
            linhas[-1] = _bordas(
                    ultima_linha,
                    detalhe,
                    largura=largura,
                    espacamento_minimo=1
                )

        for linha in linhas:
            self.texto(linha)

        if self._config.cupom.itens_modo_condensado:
            self.condensado()

        return self

    def corpo(self):
        self.corpo_I_titulo()
        self.corpo_II_consumidor()
        if not self._resumido:
            self.corpo_III_legenda()
            self.corpo_IV_itens()
        self.corpo_V_total_cupom()
        self.corpo_VI_meio_pagamento()
        self.corpo_VII_observacoes_fisco()
        self.corpo_VIII_dados_entrega()

        if self.anotacoes_antes_obs_contribuinte:
            self.normal()
            self.esquerda()
            self.avanco()
            for anotacao in self.anotacoes_antes_obs_contribuinte:
                self.texto(anotacao)

        self.corpo_IX_observacoes_contribuinte()

        if self.anotacoes_corpo:
            self.normal()
            self.esquerda()
            self.avanco()
            for anotacao in self.anotacoes_corpo:
                self.texto(anotacao)

        self.anotacoes_antes_obs_contribuinte[:] = []
        self.anotacoes_corpo[:] = []

    def corpo_I_titulo(self):
        self.normal()
        self.centro()
        self.negrito()
        self.texto('Extrato No. {}'.format(self.numero_extrato()))
        self.texto(u'CUPOM FISCAL ELETRÔNICO - SAT')
        self.indicacao_de_teste()
        self.negrito()
        self.esquerda()
        self.separador()

    def corpo_II_consumidor(self):
        # (!) Com a introdução do Anexo I, "Correlação de Campos do Extrato
        #     do CF-e-SAT", parece que ficou claro que o grupo II não deverá
        #     exibir o nome do consumidor, nem se estiver presente.
        #
        #     Como havia implementado originalmente com o nome, apenas
        #     deixei configurável a apresentação do nome do consumidor.
        #
        documento = (
                self.root.findtext('./infCFe/dest/CNPJ')
                or self.root.findtext('./infCFe/dest/CPF')
                or ''
            )

        if br.is_cnpjcpf(documento):
            self.normal()
            self.esquerda()
            self.texto('CPF/CNPJ do Consumidor: {}'.format(
                    br.as_cnpjcpf(documento)
                ))

            nome = self.root.findtext('./infCFe/dest/xNome')
            if nome and self._config.cupom.exibir_nome_consumidor:
                self.quebrar(nome)

            self.separador()

    def corpo_III_legenda(self):
        legendas = ' | '.join([
                '#', 'COD', 'DESC', 'QTD', 'UN',
                'VL UN R$', '(VL TR R$)*', 'VL ITEM R$'
            ])

        self.normal()
        self.condensado()
        self.quebrar(legendas)
        self.condensado()
        self.separador()

        self.anotacoes_antes_obs_contribuinte.append(
                u'*Valor aproximado dos tributos do item.'
            )

    def corpo_IV_itens(self):

        for det in self.root.findall('./infCFe/det'):
            self.apresentar_item(det)

            prod = det.find('prod')
            imposto = det.find('imposto')

            vProd = Decimal(prod.findtext('vProd'))  # noqa: keep for reference
            vDesc = Decimal(prod.findtext('vDesc') or 0)
            vRatDesc = Decimal(prod.findtext('vRatDesc') or 0)
            vOutro = Decimal(prod.findtext('vOutro') or 0)
            vRatAcr = Decimal(prod.findtext('vRatAcr') or 0)

            # (!) Do modo como está implementado, descontos e acréscimos
            # serão ambos tratados como se pudessem coexistir, embora o
            # Manual de Orientação seja claro quando diz que "valores de
            # descontos e acréscimos são mutuamente exclusivos".
            forcar_exibir = False

            if not vDesc.is_zero() or forcar_exibir:
                self.bordas('desconto sobre item', '- {:n}'.format(vDesc))

            if not vRatDesc.is_zero() or forcar_exibir:
                self.bordas(
                        'rateio de desconto sobre subtotal',
                        '- {:n}'.format(vRatDesc)
                    )

            if not vOutro.is_zero() or forcar_exibir:
                self.bordas(u'acréscimo sobre item', '+ {:n}'.format(vOutro))

            if not vRatAcr.is_zero() or forcar_exibir:
                self.bordas(
                        u'rateio de acréscimo sobre subtotal',
                        '+ {:n}'.format(vRatAcr)
                    )

            if imposto.find('ISSQN/cNatOp') is not None:
                # se cNatOp (U09) presente, assume item tributado pelo ISSQN
                vBC = Decimal(imposto.findtext('ISSQN/vBC') or 0)
                vDeducISSQN = Decimal(
                        imposto.findtext('ISSQN/vDeducISSQN')
                        or 0
                    )

                if not vDeducISSQN.is_zero():
                    self.bordas(
                            u'dedução para ISSQN',
                            '- {:n}'.format(vDeducISSQN)
                        )

                self.bordas(u'base de cálculo ISSQN', '{:n}'.format(vBC))

    def corpo_V_total_cupom(self):

        total = self.root.find('./infCFe/total')

        # calcula o total de descontos apenas dos itens tributados por ISSQN,
        # usando os campos vDesc (I12) e vOutro (I13);
        issqn_total_desc = ZERO
        issqn_total_acres = ZERO

        for det in self.root.findall('./infCFe/det'):
            if det.find('imposto/ISSQN/cNatOp') is not None:
                # se cNatOp (U09) presente, assume item tributado pelo ISSQN
                issqn_total_desc += Decimal(det.findtext('prod/vDesc') or 0)
                issqn_total_acres += Decimal(det.findtext('prod/vOutro') or 0)

        # total de descontos sobre itens (vDesc, W05);
        # total de outras despesas acessórias sobre itens (vOutro W10);
        total_desc = (
                issqn_total_desc
                + Decimal(total.findtext('ICMSTot/vDesc') or 0)
            )

        total_acres = (
                issqn_total_acres
                + Decimal(total.findtext('ICMSTot/vOutro') or 0)
            )

        # vProd (W04)
        vProd = Decimal(total.findtext('ICMSTot/vProd'))
        vDescSubtot = Decimal(total.findtext('DescAcrEntr/vDescSubtot') or 0)
        vAcresSubtot = Decimal(total.findtext('DescAcrEntr/vAcresSubtot') or 0)

        self.normal()
        self.avanco()

        forcar_exibir = False

        ha_desc_itens = not total_desc.is_zero() or forcar_exibir
        ha_desc_subtotal = not vDescSubtot.is_zero() or forcar_exibir
        ha_desconto = ha_desc_itens or ha_desc_subtotal or forcar_exibir

        ha_acres_itens = not total_acres.is_zero() or forcar_exibir
        ha_acres_subtotal = not vAcresSubtot.is_zero() or forcar_exibir
        ha_acrescimo = ha_acres_itens or ha_acres_subtotal or forcar_exibir

        if ha_desconto or ha_acrescimo:
            # há desconto ou acréscimo sobre os itens e/ou sobre o subtotal
            self.bordas('Total bruto de itens', '{:n}'.format(vProd))

        if ha_desc_itens:
            # há desconto sobre itens
            self.bordas(
                    'Total de descontos sobre item',
                    '- {:n}'.format(total_desc)
                )

        if ha_desc_subtotal:
            # há desconto sobre o subtotal
            self.bordas(
                    'Desconto sobre subtotal',
                    '- {:n}'.format(vDescSubtot)
                )

        if ha_acres_itens:
            # há acréscimo sobre itens
            self.bordas(
                    u'Total de acréscimos sobre item',
                    '+ {:n}'.format(total_acres)
                )

        if ha_acres_subtotal:
            # há acréscimo sobre o subtotal
            self.bordas(
                    u'Acréscimo sobre subtotal',
                    '+ {:n}'.format(vAcresSubtot)
                )

        self.negrito()
        self.bordas('TOTAL R$', '{:n}'.format(Decimal(total.findtext('vCFe'))))
        self.negrito()

    def corpo_VI_meio_pagamento(self):
        self.normal()
        self.avanco()

        for mp in self.root.findall('./infCFe/pgto/MP'):
            self.bordas(
                    ersat.meio_pagamento(mp.findtext('cMP')),
                    '{:n}'.format(Decimal(mp.findtext('vMP')))
                )

        valor_troco = Decimal(self.root.findtext('./infCFe/pgto/vTroco') or 0)
        if not valor_troco.is_zero():
            self.bordas('Troco R$', '{:n}'.format(valor_troco))

    def corpo_VII_observacoes_fisco(self):
        """
        Grupo VII, Observações do Fisco. Redação atual, efeitos até 31-12-2016.
        Usa o atributo ``xCampo`` (Z04) e o elemento ``xTexto`` (Z05):

        .. sourcecode:: xml

            <CFe>
              <infCFe>
                <infAdic>
                  <obsFisco xCampo="Foo">
                      <xTexto>Conteúdo de Foo</xTexto>
                  </obsFisco>
                  <obsFisco xCampo="Bar">
                      <xTexto>Conteúdo de Bar</xTexto>
                  </obsFisco>
                </infAdic>
              </infCFe>
            </CFe>

        Nova redação, efeitos a partir de 01-01-2017.
        Usa o atributo ``xCampo`` (ZA02) e o elemento ``xTexto`` (ZA03):

        .. sourcecode:: xml

            <CFe>
              <infCFe>
                <obsFisco xCampo="Spam">
                    <xTexto>Conteúdo de Spam</xTexto>
                </obsFisco>
                <obsFisco xCampo="Eggs">
                    <xTexto>Conteúdo de Eggs</xTexto>
                </obsFisco>
              </infCFe>
            </CFe>

        """
        iniciado = False

        for obs in self.root.findall('./infCFe/infAdic/obsFisco'):
            if not iniciado:
                self.normal()
                self.esquerda()
                self.avanco()
                self.condensado()
                iniciado = True

            self.quebrar(u'{}: {}'.format(
                    obs.attrib['xCampo'],
                    obs.findtext('xTexto')))

        if iniciado:
            self.condensado()

    def corpo_VIII_dados_entrega(self):

        entrega = self.root.find('./infCFe/entrega')
        if entrega is None:
            return

        logradouro = entrega.findtext('xLgr')
        numero = entrega.findtext('nro')
        complemento = entrega.findtext('xCpl')
        bairro = entrega.findtext('xBairro')

        if numero:  # número não é obrigatório
            # mas existe, então o coloca próximo ao logradouro
            logradouro = u'{}, {}'.format(logradouro, numero)

        if complemento:  # complemento não é obrigatório
            # faz um esforço para manter o complemento na mesma linha que o
            # logradouro/número se couber, senão o complemento irá usar uma
            # linha exclusiva...
            if len(logradouro) + len(complemento) < self._colunas:
                # ok, mantém o complemento na mesma linha que o logradouro
                logradouro = u'{}, {}'.format(logradouro, complemento)
                complemento = ''  # ignora a linha que deveria conter o xCpl

        cidade = u'{}/{}'.format(
                entrega.findtext('xMun'),
                entrega.findtext('UF')
            )

        partes_endereco = [logradouro, complemento, bairro, cidade]
        endereco = '\r\n'.join([e for e in partes_endereco if e])

        self.normal()
        self.esquerda()
        self.separador()
        self.negrito()
        self.texto('DADOS PARA ENTREGA')
        self.negrito()
        self.quebrar(u'Endereço: {}'.format(endereco))

        nome_destinatario = self.root.findtext('./infCFe/dest/xNome')
        if nome_destinatario:
            self.quebrar(u'Destinatário: {}'.format(nome_destinatario))

    def corpo_IX_observacoes_contribuinte(self):
        infCpl = self.root.findtext('./infCFe/infAdic/infCpl')
        vCFeLei12741 = Decimal(
                self.root.findtext('./infCFe/total/vCFeLei12741')
                or 0
            )

        if infCpl or not vCFeLei12741.is_zero():
            self.normal()
            self.esquerda()
            self.separador()
            self.negrito()
            self.texto(u'OBSERVAÇÕES DO CONTRIBUINTE')
            self.negrito()

            if infCpl:
                self.condensado()
                self.quebrar(infCpl)
                self.condensado()

            if not vCFeLei12741.is_zero():
                self.condensado()
                self.quebrar(u'Valor aproximado dos tributos deste cupom')
                self.bordas(
                        u'(Lei Fed. 12.741/2012) R$',
                        '{:n}'.format(vCFeLei12741)
                    )
                self.condensado()

    def rodape(self):

        infCFe = self.root.find('./infCFe')
        sat_numero_serie = 'SAT no. {}'.format(
                infCFe.findtext('ide/nserieSAT')
            )

        datahora = datetime.strptime('{}{}'.format(
                infCFe.findtext('ide/dEmi'),
                infCFe.findtext('ide/hEmi')), '%Y%m%d%H%M%S')

        datahora_emissao = datahora.strftime('%d/%m/%Y - %H:%M:%S')
        if six.PY2:
            # Python 2: strftime() resulta em str, que precisará ser passado
            # como um objeto unicode para unidecode (em self.texto())
            datahora_emissao = datahora_emissao.decode('utf-8')

        chave = ersat.ChaveCFeSAT(infCFe.attrib['Id'])

        self.normal()
        self.separador()
        self.centro()
        self.negrito()
        self.texto(sat_numero_serie)
        self.negrito()
        self.texto(datahora_emissao)

        self.avanco()
        self.condensado()
        self.texto(' '.join(chave.partes()))
        self.condensado()

        self.avanco()

        self.chave_cfe_code128(chave)

        self.avanco(2)
        self.impressora.qrcode(
                ersat.dados_qrcode(self._tree),
                qrcode_module_size=self._config.qrcode.tamanho_modulo,
                qrcode_ecc_level=self._config.qrcode.nivel_correcao
            )

        self.avanco()
        self.qrcode_mensagem()



# -*- coding: utf-8 -*-
#
# setup.py
#
# Copyright 2015 Base4 Sistemas Ltda ME
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import
from __future__ import unicode_literals

import io
import os
import re

from setuptools import setup


def read(*filenames, **kwargs):
    encoding = kwargs.get('encoding', 'utf-8')
    sep = kwargs.get('sep', '\n')
    buf = []
    for filename in filenames:
        with io.open(filename, 'r', encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


def read_version():
    content = read(os.path.join('satextrato', '__init__.py'))
    return re.search(r"__version__ = '([^']+)'", content).group(1)


long_description = read('README.rst')


setup(
        name='satextrato',
        version=read_version(),
        description=u'Impressão dos Extratos do CF-e-SAT',
        long_description=long_description,
        long_description_content_type='text/x-rst',
        packages=[
                'satextrato',
            ],
        install_requires=[
                'six',
                'unidecode',
                'satcomum',
                'PyESCPOS>=0.3',
            ],
        include_package_data=True,
        license='Apache Software License',
        platforms='any',
        url='http://github.com/base4sistemas/satextrato/',
        author=u'Daniel Gonçalves',
        author_email='daniel@base4.com.br',
        classifiers=[
                'Development Status :: 4 - Beta',
                'Environment :: Other Environment',
                'Intended Audience :: Developers',
                'Intended Audience :: Information Technology',
                'License :: OSI Approved :: Apache Software License',
                'Natural Language :: Portuguese (Brazilian)',
                'Operating System :: OS Independent',
                'Programming Language :: Python',
                'Programming Language :: Python :: 2.7',
                'Programming Language :: Python :: 3.6',
                'Programming Language :: Python :: 3.7',
                'Programming Language :: Python :: 3.8',
                'Topic :: Office/Business :: Financial :: Point-Of-Sale',
                'Topic :: Printing',
                'Topic :: Software Development :: Libraries :: Python Modules',
                'Topic :: Utilities',
            ]
    )



# -*- coding: utf-8 -*-
#
# satextrato/config.py
#
# Copyright 2015 Base4 Sistemas Ltda ME
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import absolute_import
from __future__ import unicode_literals

import io
import os
import sys

from collections import namedtuple

try:
    # http://python-future.org/compatible_idioms.html#configparser
    # Python 2 and 3 (after ``pip install configparser``)
    from configparser import ConfigParser
except ImportError:
    # fallback to Python 2 SafeConfigParser module
    from ConfigParser import SafeConfigParser as ConfigParser


SECAO_CUPOM = 'cupom'
SECAO_RODAPE = 'rodape'
SECAO_CODE128 = 'code128'
SECAO_QRCODE = 'qrcode'

_TAMANHO_CHAVE_CFESAT = 44


Cupom = namedtuple('Cupom', [
        'itens_modo_condensado',  # bool
        'exibir_nome_consumidor',  # bool
        'avancar_linhas',  # int
        'cortar_documento',  # bool
        'cortar_parcialmente',  # bool
        'cortar_avanco',  # int
    ])

Rodape = namedtuple('Rodape', [
        'esquerda',  # str
        'direita',  # str
    ])

Code128 = namedtuple('Code128', [
        'ignorar',  # bool
        'altura',  # int
        'quebrar',  # bool
        'quebrar_partes',  # tuple(int, ...)
        'truncar',  # bool
        'truncar_tamanho',  # int
    ])

QRCode = namedtuple('QRCode', [
        'tamanho_modulo',  # int
        'nivel_correcao',  # str
        'nome_aplicativo',  # str
        'mensagem_modo_condensado',  # bool
        'mensagem',  # str
    ])

Configuracoes = namedtuple('Configuracoes', [
        'cupom',  # Cupom
        'rodape',  # Rodape
        'code128',  # Code128
        'qrcode',  # QRCode
    ])


def carregar(
        arquivo=None,
        secao_cupom=SECAO_CUPOM,
        secao_rodape=SECAO_RODAPE,
        secao_code128=SECAO_CODE128,
        secao_qrcode=SECAO_QRCODE
        ):
    """Carrega as configurações do extrato do CF-e-SAT.

    :param str arquivo: Opcional. Caminho completo para o arquivo de
        onde serão carregadas as configurações. O arquivo será interpretado
        pela classe ``ConfigParser`` da biblioteca padrão.
        Se arquivo não existir, será criado um com os valores padrão.
        Se um nome de arquivo não for informado, as configurações serão
        carregadas do caminho e arquivo definidos nas variáveis de ambiente
        ``SATEXTRATO_CONFIG_DIR`` e ``SATEXTRATO_CONFIG_FILENAME``.

    :param str secao_cupom: Opcional. Nome da seção onde ficam as
        propriedades associadas ao cupom do extrato. Veja :attr:`Cupom`.

    :param str secao_rodape: Opcional. Nome da seção onde ficam as
        propriedades associadas ao rodapé do extrato. Veja :attr:`Rodape`.

    :param str secao_code128: Opcional. Nome da seção onde ficam as
        propriedades associadas ao código de barras padrão Code-128 que é
        impresso no extrato. Veja :attr:`Code128`.

    :param str secao_qrcode: Opcional. Nome da seção onde ficam as
        propriedades associadas ao QRCode que é impresso no extrato.
        Veja :attr:`QRCode`.

    :rtype: satextrato.config.Configuracoes

    """
    arquivo = arquivo or os.path.join(_config_dir(), _config_filename())
    if not os.path.isfile(arquivo):
        salvar(
                padrao(),
                arquivo=arquivo,
                secao_cupom=secao_cupom,
                secao_rodape=secao_rodape,
                secao_code128=secao_code128,
                secao_qrcode=secao_qrcode
            )

    parser = ConfigParser()

    with io.open(arquivo, 'r', encoding='utf-8') as f:
        if hasattr(parser, 'read_file'):
            parser.read_file(f)
        else:
            # fallback to Python 2 ConfigParser.readfp() (deprecated)
            parser.readfp(f)

    cupom = Cupom(
            itens_modo_condensado=parser.getboolean(
                    secao_cupom,
                    'itens_modo_condensado'
                ),
            exibir_nome_consumidor=parser.getboolean(
                    secao_cupom,
                    'itens_modo_condensado'
                ),
            avancar_linhas=parser.getint(secao_cupom, 'avancar_linhas'),
            cortar_documento=parser.getboolean(
                    secao_cupom,
                    'itens_modo_condensado'
                ),
            cortar_parcialmente=parser.getboolean(
                    secao_cupom,
                    'itens_modo_condensado'
                ),
            cortar_avanco=parser.getint(secao_cupom, 'cortar_avanco')
        )

    rodape = Rodape(
            esquerda=parser.get(secao_rodape, 'esquerda'),
            direita=parser.get(secao_rodape, 'direita')
        )

    code128 = Code128(
            ignorar=parser.getboolean(secao_code128, 'ignorar'),
            altura=parser.getint(secao_code128, 'altura'),
            quebrar=parser.getboolean(secao_code128, 'quebrar'),
            quebrar_partes=parser.get(secao_code128, 'quebrar_partes'),
            truncar=parser.getboolean(secao_code128, 'truncar'),
            truncar_tamanho=parser.getint(secao_code128, 'truncar_tamanho')
        )

    qrcode = QRCode(
        tamanho_modulo=parser.getint(secao_qrcode, 'tamanho_modulo'),
        nivel_correcao=parser.get(secao_qrcode, 'nivel_correcao'),
        nome_aplicativo=parser.get(secao_qrcode, 'nome_aplicativo'),
        mensagem=parser.get(secao_qrcode, 'mensagem'),
        mensagem_modo_condensado=parser.getboolean(
                secao_qrcode,
                'mensagem_modo_condensado'
            )
        )

    if sys.version_info >= (3,):
        # Python 3, converte o valor do nível de correção para byte
        qrcode = qrcode._replace(
                nivel_correcao=qrcode.nivel_correcao.encode('ascii')
            )

    partes = _validar_code128_quebrar_partes(code128.quebrar_partes)
    code128 = code128._replace(quebrar_partes=partes)

    conf = Configuracoes(
            cupom=cupom,
            rodape=rodape,
            code128=code128,
            qrcode=qrcode
        )

    return conf


def salvar(
        conf,
        arquivo=None,
        secao_cupom=SECAO_CUPOM,
        secao_rodape=SECAO_RODAPE,
        secao_code128=SECAO_CODE128,
        secao_qrcode=SECAO_QRCODE
        ):
    """Salva configurações do extrato do CF-e-SAT.

    :param conf: Instância de :attr:`Configuracoes` (``namedtuple``).

    :param str arquivo: Opcional. Caminho completo para o arquivo onde
        serão salvas as configurações. Se não informado, será salvo no
        caminho e arquivo definidos nas variáveis de ambiente
        ``SATEXTRATO_CONFIG_DIR`` e ``SATEXTRATO_CONFIG_FILENAME``.

    :param str secao_cupom: Opcional. Nome da seção onde ficam as
        propriedades associadas ao cupom do extrato. Veja :attr:`Cupom`.

    :param str secao_rodape: Opcional. Nome da seção onde ficam as
        propriedades associadas ao rodapé do extrato. Veja :attr:`Rodape`.

    :param str secao_code128: Opcional. Nome da seção onde ficam as
        propriedades associadas ao código de barras padrão Code-128 que é
        impresso no extrato. Veja :attr:`Code128`.

    :param str secao_qrcode: Opcional. Nome da seção onde ficam as
        propriedades associadas ao QRCode que é impresso no extrato.
        Veja :attr:`QRCode`.

    """
    parser = ConfigParser()

    sec = _SectionHelper(secao_cupom, parser, conf.cupom)
    sec.setboolean('itens_modo_condensado')
    sec.setboolean('exibir_nome_consumidor')
    sec.setint('avancar_linhas')
    sec.setboolean('cortar_documento')
    sec.setboolean('cortar_parcialmente')
    set.setint('cortar_avanco')

    sec = _SectionHelper(secao_rodape, parser, conf.rodape)
    sec.set('esquerda')
    sec.set('direita')

    # converte a partir de uma tupla de inteiros
    partes = ','.join(str(p) for p in conf.code128.quebrar_partes)

    sec = _SectionHelper(secao_code128, parser, conf.code128)
    sec.setboolean('ignorar')
    sec.setint('altura')
    sec.setboolean('quebrar')
    sec.setvalue('quebrar_partes', partes)
    sec.setboolean('truncar')
    sec.setint('truncar_tamanho')

    qr_level = conf.qrcode.nivel_correcao  # no Python 3 é do tipo byte
    if sys.version_info >= (3,):
        # converte de byte para str no Python 3
        qr_level = qr_level.decode('utf-8')

    sec = _SectionHelper(secao_qrcode, parser, conf.qrcode)
    sec.setint('tamanho_modulo')
    sec.setvalue('nivel_correcao', qr_level)
    sec.set('nome_aplicativo')
    sec.set('mensagem')
    sec.setboolean('mensagem_modo_condensado')

    arquivo = arquivo or os.path.join(_config_dir(), _config_filename())
    with io.open(arquivo, 'w', encoding='utf-8') as f:
        parser.write(f)


class _SectionHelper(object):

    def __init__(self, section_name, parser, obj):
        self._section_name = section_name
        self._parser = parser
        self._obj = obj

    def _set(self, option_name, value):
        return self._parser.set(self._section_name, option_name, value)

    def set(self, option_name):
        return self._set(option_name, getattr(self._obj, option_name))

    def setboolean(self, option_name):
        value = 'yes' if getattr(self._obj, option_name) else 'no'
        return self._set(option_name, value)

    def setint(self, option_name):
        value = str(getattr(self._obj, option_name))
        return self._set(option_name, value)

    def setvalue(self, option_name, value):
        return self._parser.set(self._section_name, option_name, value)


def _config_dir():
    default_path = os.path.join(os.path.expanduser('~'), '.satextrato')
    path = os.getenv('SATEXTRATO_CONFIG_DIR', default_path)
    if not os.path.isdir(path):
        os.makedirs(path)
    return path


def _config_filename():
    return os.getenv('SATEXTRATO_CONFIG_FILENAME', 'satextrato.ini')


def padrao():
    from escpos import barcode

    cupom = Cupom(
            itens_modo_condensado=True,
            exibir_nome_consumidor=False,
            avancar_linhas=7,
            cortar_documento=True,
            cortar_parcialmente=True,
            cortar_avanco=0
        )

    rodape = Rodape(
            esquerda='Extrato CF-e-SAT',
            direita='http://git.io/vJRRk'
        )

    code128 = Code128(
            ignorar=False,
            altura=56,
            quebrar=True,
            quebrar_partes=(22, 22),
            truncar=False,
            truncar_tamanho=44
        )

    qrcode = QRCode(
            tamanho_modulo=barcode.QRCODE_MODULE_SIZE_4,
            nivel_correcao=barcode.QRCODE_ERROR_CORRECTION_L,
            nome_aplicativo='De Olho Na Nota',
            mensagem_modo_condensado=True,
            mensagem=(
                    u'Consulte o QRCode pelo aplicativo De Olho Na Nota, '
                    u'disponível na AppStore (Apple) e PlayStore (Android)'
                )
        )

    conf = Configuracoes(
            cupom=cupom,
            rodape=rodape,
            code128=code128,
            qrcode=qrcode
        )

    return conf


def _validar_code128_quebrar_partes(partes):
    # <partes> str - uma lista de inteiros separados por vírgulas
    try:
        lista_partes = [int(p) for p in partes.split(',')]
    except ValueError:
        raise ValueError(
                (
                    'Configuracoes do extrato do CF-e-SAT, Code128 em '
                    'partes deve especificar as partes em valores inteiros, '
                    'todos numeros pares e separados por virgulas; '
                    'obtido: {!r}'
                ).format(partes)
            )
    else:
        calculado = sum(lista_partes)
        if calculado != _TAMANHO_CHAVE_CFESAT:
            raise ValueError(
                (
                    'Configuracoes do extrato do CF-e-SAT, Code128 em '
                    'partes deve especificar as partes em valores inteiros, '
                    'todos numeros pares e separados por virgulas; a soma '
                    'das partes deve ser igual ao tamanho da chave do '
                    'CF-e-SAT; obtido: {!r} (esperado: {!r})'
                ).format(
                    calculado,
                    _TAMANHO_CHAVE_CFESAT
                )
            )

    # todos os números inteiros em <partes> devem ser pares
    for i, n in enumerate(lista_partes, 1):
        if n <= 0 or n % 2 != 0:
            raise ValueError(
                (
                    'Configuracoes do extrato do CF-e-SAT, Code128, '
                    'elemento {!r} deve ser um número par; obtido {!r}'
                ).format(i, n)
            )

    return tuple(lista_partes)
